@@ -7,7 +7,7 @@
 //!
 //! ### Loopback links
 //!
-//! Write an expression to a loopback link, and then read it back from the same link
+//! Write an expression to a loopback [`Link`], and then read it back from the same link
 //! object:
 //!
 //! ```
@@ -26,7 +26,7 @@
 //! let mut buffer = String::new();
 //!
 //! for _ in 0 .. link.test_head("System`List")? {
-//!     buffer.push_str(link.get_string_ref()?.to_str())
+//!     buffer.push_str(link.get_string_ref()?.as_str())
 //! }
 //!
 //! assert_eq!(buffer, "abc");
@@ -104,8 +104,9 @@
 //! is built on top of the [WSTP C API][CFunctions].
 //!
 //! When using the `wstp` crate as a dependency, the `wstp` crate's cargo build script
-//! will use `wolfram-app-discovery` to automatically locate any local installations of
-//! the Wolfram Language, and will link against the WSTP static library located within.
+//! will use [`wolfram-app-discovery`][wolfram-app-discovery] to automatically find any
+//! local installations of the Wolfram Language, and will link against the WSTP static
+//! library located within.
 //!
 //! The [Wolfram Engine][WolframEngine] can be downloaded and used for free for
 //! non-commercial or pre-production uses. A license must be purchased when used as part
@@ -130,6 +131,8 @@
 //! [WolframEngine]: https://www.wolfram.com/engine/
 //! [WE-FAQ]: https://www.wolfram.com/engine/faq/
 //! [CFunctions]: https://reference.wolfram.com/language/guide/AlphabeticalListingOfWSTPCFunctions.html
+//!
+//! [wolfram-app-discovery]: https://crates.io/crates/wolfram-app-discovery
 
 #![warn(missing_docs)]
 
@@ -144,6 +147,8 @@ mod put;
 
 mod strx;
 
+pub mod kernel;
+
 /// Ensure that doc tests in the README.md file get run.
 #[doc(hidden)]
 mod test_readme {
@@ -170,8 +175,9 @@ use wstp_sys::{WSErrorMessage, WSReady, WSReleaseErrorMessage, WSLINK};
 pub use wstp_sys as sys;
 
 pub use crate::{
+    env::shutdown,
     error::Error,
-    get::{Array, LinkStr},
+    get::{Array, LinkStr, Token, TokenType},
     link_server::LinkServer,
     strx::{Ucs2Str, Utf16Str, Utf32Str, Utf8Str},
 };
@@ -650,7 +656,69 @@ impl Link {
 
     /// Read an expression off of this link.
     pub fn get_expr(&mut self) -> Result<Expr, Error> {
-        get_expr(self)
+        self.get_expr_with_resolver(&mut |_| None)
+    }
+
+    // TODO: This needs a bit more design work before being made public. For starters,
+    //       you have to pass a closure to it using `get_expr_with_resolver(&mut |_| ...)`
+    //       which looks out of place. Using `dyn FnMut()` is to avoid having to
+    //       monomorphize different copies of `get_expr_with_resolver()`
+    #[doc(hidden)]
+    pub fn get_expr_with_resolver(
+        &mut self,
+        mut resolver: &mut dyn FnMut(&str) -> Option<Symbol>,
+    ) -> Result<Expr, Error> {
+        let value = self.get_token()?;
+
+        let expr: Expr = match value {
+            Token::Integer(value) => Expr::from(value),
+            Token::Real(value) => {
+                let real: wolfram_expr::F64 = match wolfram_expr::F64::new(value) {
+                    Ok(real) => real,
+                    // TODO: Try passing a NaN value or a BigReal value through WSLINK.
+                    Err(_is_nan) => {
+                        return Err(Error::custom(format!(
+                        "NaN value passed on WSLINK cannot be used to construct an Expr"
+                    )))
+                    },
+                };
+                Expr::number(Number::Real(real))
+            },
+            Token::String(value) => Expr::string(value.as_str()),
+            Token::Symbol(value) => {
+                let symbol_str: &str = value.as_str();
+
+                // If `symbol_str` is not an absolute symbol, use the provided `resolver`
+                // to attempt to resolve it into a concrete Symbol.
+                let symbol = Symbol::try_new(symbol_str).or_else(|| resolver(symbol_str));
+
+                let symbol: Symbol = match symbol {
+                    Some(sym) => sym,
+                    None => {
+                        return Err(Error::custom(format!(
+                            "symbol name '{}' has no context",
+                            symbol_str
+                        )))
+                    },
+                };
+
+                Expr::symbol(symbol)
+            },
+            Token::Function { length: arg_count } => {
+                drop(value);
+
+                let head = self.get_expr_with_resolver(&mut resolver)?;
+
+                let mut contents = Vec::with_capacity(arg_count);
+                for _ in 0..arg_count {
+                    contents.push(self.get_expr_with_resolver(&mut resolver)?);
+                }
+
+                Expr::normal(head, contents)
+            },
+        };
+
+        Ok(expr)
     }
 
     /// Write an expression to this link.
@@ -747,68 +815,6 @@ impl Link {
     }
 }
 
-//======================================
-// Read from the link
-//======================================
-
-fn get_expr(link: &mut Link) -> Result<Expr, Error> {
-    use wstp_sys::{WSTKFUNC, WSTKINT, WSTKREAL, WSTKSTR, WSTKSYM};
-
-    let type_: i32 = link.get_raw_type()?;
-
-    let expr: Expr = match type_ as u8 {
-        WSTKINT => Expr::from(link.get_i64()?),
-        WSTKREAL => {
-            let real: wolfram_expr::F64 = match wolfram_expr::F64::new(link.get_f64()?) {
-                Ok(real) => real,
-                // TODO: Try passing a NaN value or a BigReal value through WSLINK.
-                Err(_is_nan) => {
-                    return Err(Error::custom(format!(
-                        "NaN value passed on WSLINK cannot be used to construct an Expr"
-                    )))
-                },
-            };
-            Expr::number(Number::Real(real))
-        },
-        WSTKSTR => Expr::string(link.get_string_ref()?.as_str()),
-        WSTKSYM => {
-            let symbol_link_str = link.get_symbol_ref()?;
-            let symbol_str = symbol_link_str.as_str();
-
-            let symbol: Symbol = match Symbol::try_new(symbol_str) {
-                Some(sym) => sym,
-                None => {
-                    return Err(Error::custom(format!(
-                        "symbol name '{}' has no context",
-                        symbol_str
-                    )))
-                },
-            };
-
-            Expr::symbol(symbol)
-        },
-        WSTKFUNC => {
-            let arg_count = link.get_arg_count()?;
-
-            let head = link.get_expr()?;
-
-            let mut contents = Vec::with_capacity(arg_count);
-            for _ in 0..arg_count {
-                contents.push(link.get_expr()?);
-            }
-
-            Expr::normal(head, contents)
-        },
-        _ => return Err(Error::custom(format!("unknown WSLINK type: {}", type_))),
-    };
-
-    Ok(expr)
-}
-
-//======================================
-// Write to the link
-//======================================
-
 //======================================
 // Utilities
 //======================================
